/* partialy inspired by smbmnt.c */

#include <stdlib.h>
#include <string.h>
#include <stdio.h>
#include <errno.h>
#include <unistd.h>
#include <pwd.h>
#include <grp.h>
#include <fcntl.h>
#include <mntent.h>
#include <sys/utsname.h>
#include <sys/mount.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <sys/stat.h>
#include <sys/param.h>

#define SHELL		"/bin/sh"
#define CMD_MAX		1024
#define EXEC_MAX	20

#define SHFS_VERSION	1

#if DEBUG_LEVEL >= 3
#	define MOUNT_DEBUG
#endif

static void usage(void)
{
	fprintf(stderr, "Usage: shfsmount [opts] [user@]host[:root] mount_point [-o mount_opts]\n\n"
		        "Mount remote file system using shell session.\n\n"
			"opts:\n"
			"  -r\t\tUse rsh (instead of ssh).\n"
			"  -u user\tExecute ssh as this user (root only)\n"
			"  -P port\tConnect to this port (ssh only).\n"
			"  -O options\tSpecify ssh option (e.g. -O \"-c 3des\" == ssh -c 3des)\n"
			"  -v\t\tPrint version number.\n"
			"  user\t\tLog in using this user name.\n"
			"  host\t\tSpecify remote host.\n"
			"  root\t\tMake this directory root for mounted filesystem.\n\n"
			"mount_opts (separated by comma):\n"
			"  nocache\tDisable read-ahead and write-back cache.\n"
			"  generic\tDisable use optimized shell command set.\n"
			"  stable\tDerefference symbolic links (if possible).\n"
			"  preserve\tPreserve uid/gid (root only, implies ro).\n"
			"  ttl=sec\tTime to live for directory cache.\n"
			"  uid=user\tOwner of all files/dirs on mounted filesystem (root only).\n"
			"  gid=group\tGroup of all files/dirs on mounted filesystem (root only).\n"
			"  rmode=mode\tRoot dir mode.\n"
			"  suid, dev\tsee mount(8) (root only)\n"
			"  ro, rw, nosuid, nodev, exec, noexec\t see mount(8)\n\n"
			"Report bugs to: qiq@atrey.karlin.mff.cuni.cz\n");
	exit(1);
}

static char *fullpath(char *path)
{
        char full[MAXPATHLEN];

	if (strlen(path) > MAXPATHLEN-1)
		return NULL;
        if (!realpath(path, full))
		return NULL;
	return strdup(full);
}

static int mount_ok(char *dir)
{
	struct stat st;

	if (chdir(dir) != 0)
		return 0;
        if (lstat(".", &st) != 0)
		return 0;
        if (!S_ISDIR(st.st_mode)) {
                errno = ENOTDIR;
                return 0;
        }
        if ((getuid() != 0) && 
	    ((getuid() != st.st_uid) || 
	     ((st.st_mode & S_IRWXU) != S_IRWXU))) {
                errno = EPERM;
                return 0;
        }
        return 1;
}

static int update_mtab(char *mnt, char *dev, char *opts)
{
	int fd;
	FILE *mtab;
	struct mntent ment;

        ment.mnt_fsname = dev;
        ment.mnt_dir = mnt;
        ment.mnt_type = "shfs";
        ment.mnt_opts = opts;
        ment.mnt_freq = 0;
        ment.mnt_passno = 0;

        if ((fd = open(MOUNTED"~", O_RDWR|O_CREAT|O_EXCL, 0600)) == -1) {
                fprintf(stderr, "Can't get "MOUNTED"~ lock file");
                return 0;
        }
        close(fd);
        if ((mtab = setmntent(MOUNTED, "a+")) == NULL) {
                fprintf(stderr, "Can't open " MOUNTED);
                return 0;
        }
        if (addmntent(mtab, &ment) == 1) {
                fprintf(stderr, "Can't write mount entry");
                return 0;
        }
        if (fchmod(fileno(mtab), 0644) == -1) {
                fprintf(stderr, "Can't set perms on "MOUNTED);
                return 0;
        }
        endmntent(mtab);
        if (unlink(MOUNTED"~") == -1) {
                fprintf(stderr, "Can't remove "MOUNTED"~");
                return 0;
        }
	return 1;
}


static uid_t get_uid(char *s)
{
	struct passwd *passwd;
	char *e;
	long id = strtol(s, &e, 10);

	if (*s && *e == '\0')
		passwd = getpwuid(id);
	else
		passwd = getpwnam(s);
	if (!passwd) {
		fprintf(stderr, "Unknown user: %s\n", s);
		exit(1);
	}

	return passwd->pw_uid;
}

static gid_t get_gid(char *s)
{
	struct group *group;
	char *e;
	long id = strtol(s, &e, 10);

	if (*s && *e == '\0')
		group = getgrgid(id);
	else
		group = getgrnam(s);
	if (!group) {
		fprintf(stderr, "Unknown group: %s\n", s);
		exit(1);
	}
	
	return group->gr_gid;
}

int main(int argc, char **argv)
{
	int use_rsh = 0;
	char *root = NULL;
	char *mnt;
	char *host;
	char *long_host;
	uid_t uid = 0;
	int have_uid = 0;
	char *execv[EXEC_MAX+10];
	char opt[CMD_MAX];
	char cmd[CMD_MAX];
	char *c;
	int i, j, res;
	int null;
	int shin[2];
	int shout[2];

	if (argc < 2)
		usage();	

	execv[0] = "ssh";
	j = 1;			/* argv index for ssh/rsh */
	for (i = 1; i < argc; i++) {
		if (!strcmp(argv[i], "-v")) {
			fprintf(stderr, "shfsmount version %s\n", VERSION);
			exit(0);
		} else if (!strcmp(argv[i], "-r")) {
			execv[0] = "rsh";
			use_rsh = 1;
		} else if ((i+1 < argc) && !strcmp(argv[i], "-P")) {
			execv[j++] = "-p";
			execv[j++] = argv[i+1];
			i++;
		} else if ((i+1 < argc) && !strcmp(argv[i], "-O")) {
			char *token;

			token = strtok(argv[i+1], " \t\n");
			while (j < EXEC_MAX && token) {
				execv[j++] = token;
				token = strtok(NULL, " \t\n");
			}
			i++;
		} else if ((i+1 < argc) && !strcmp(argv[i], "-u")) {
			if (getuid() != 0) {
				fprintf(stderr, "Only root can do this!\n");
				exit(1);
			}
			uid = get_uid(argv[i+1]);
			have_uid = 1;
			i++;
		} else
			break;
	}

	if (i+1 >= argc)
		usage();
		
	host = argv[i++];
	long_host = strdup(host);
        if ((c = strchr(host, '@'))) {
                *c = '\0';
		execv[j++] = "-l";
		execv[j++] = host;
                host = c + 1;
	}

       /*
	* stuart@gnqs.org
	* Updated 1st June 2003
	*
	* the colon is now used as the separator
	* between the hostname and the root dir
	*
	* this makes it work more like scp(1)
	*/

	if ((c = strchr(host, ':'))) {
                *c = '\0';
		root = c + 1;
	}

	mnt = fullpath(argv[i++]);
	if (!mnt) {
		fprintf(stderr, "Invalid path: %s\n", argv[i-1]);
		exit(1);
	}
	if (!mount_ok(mnt)) {
		fprintf(stderr, "%s: %s\n", mnt, strerror(errno));
		exit(1);
	}

	opt[0] = '\0';
	snprintf(opt, sizeof(opt), ",version=%d,mnt=%.256s", SHFS_VERSION, mnt);
	if (root) {
		strncat(opt, ",root=", sizeof(opt)-strlen(opt)-1);
		strncat(opt, root, sizeof(opt)-strlen(opt)-1);
	}

	if (i < argc) {
			char *s;
		if (i+1 < argc && !strcmp(argv[i], "-o")) {
			char r[15];
			
			s = argv[i+1];
			while (s) {
				if ((c = strpbrk(s, ",")))
					*(c++) = '\0';
					
				/* check for root options */
				if ((getuid() != 0) && !(strcmp(s, "preserve") 
				   && strncmp(s, "uid=", 4) && strncmp(s, "gid=", 4) 
				   && strcmp(s, "suid") && strcmp(s, "dev"))) {
					fprintf(stderr, "'%s' used by non-root user!\n", s);
					exit(1);
				}
				
				if (!strncmp(s, "uid=", 4)) {
					snprintf(r, sizeof(r), "uid=%u", get_uid(s+4));
					s = r;
				}
				if (!strncmp(s, "gid=", 4)) {
					snprintf(r, sizeof(r), "gid=%u", get_gid(s+4));
					s = r;
				}

				if (strlen(opt)+1+strlen(s)+1 < sizeof(opt)) {
					strcat(opt, ",");
					strcat(opt, s);
				}
				s = c;
			}
		} else {
			fprintf(stderr, "%s: invalid option\n", argv[i]);
			exit(1);
		}
	}

	/* ensure fd 0-2 are open */
	do {
		if ((null = open("/dev/null", 0)) < 0) {
			fprintf(stderr, "open(/dev/null): %s\n", strerror(errno));
			exit(1);
		}
	} while (null < 3);
	close(null);
#undef USE_PIPES
#ifdef USE_PIPES
	if ((pipe(shin) == -1) || (pipe(shout) == -1)) {
		fprintf(stderr, "pipe: %s\n", strerror(errno));
		exit(1);
	}
#else
	if (socketpair(AF_UNIX, SOCK_STREAM, 0, shin) == -1) {
		fprintf(stderr, "socketpair: %s\n", strerror(errno));
		exit(1);
	}
	shout[0] = dup(shin[1]);
	shout[1] = dup(shin[0]);
	if (shout[0] == -1 || shout[1] == -1) {
		fprintf(stderr, "dup: %s\n", strerror(errno));
		exit(1);
	}
#endif

	switch (fork()) {
	case 0:			 /* child */
		if ((dup2(shin[0], STDIN_FILENO) == -1) || 
		    (dup2(shout[1], STDOUT_FILENO) == -1)) {
			fprintf(stderr, "dup2: %s\n", strerror(errno));
			exit(1);
		}
		close(shin[0]);
		close(shin[1]);
		close(shout[1]);
		close(shout[0]);
		close(2);		/* stderr closed because of autofs */
		
		execv[j++] = host;
		execv[j++] = SHELL;
		execv[j] = NULL;

		if (!have_uid)
			uid = getuid();
		if (setuid(uid)) {
			fprintf(stderr, "setuid: %s\n", strerror(errno));
			exit(1);
		}

		if (use_rsh)
			execvp("rsh", execv);
		else
			execvp("ssh", execv);

		fprintf(stderr, "exec: %s: %s\n", execv[0], strerror(errno));
		exit(1);
	case -1:
		fprintf(stderr, "fork: %s\n", strerror(errno));
		res = 1;
		goto out;
	}
	
	close(shin[0]);
	close(shout[1]);
	
	snprintf(cmd, sizeof(cmd), "fdi=%d,fdo=%d%s", shout[0], shin[1], opt);
	if (strlen(cmd) >= sizeof(cmd)-1) {
		fprintf(stderr, "%s: Agrs too long!\n", argv[0]);
		res = 1;
		goto out;
	}
	
#ifdef MOUNT_DEBUG
	printf("uid: %d, euid: %d\n", getuid(), geteuid());
	printf("%s\n", cmd);
#endif
	if (mount("none", mnt, "shfs", 0, cmd) < 0) {
		if (errno == ENODEV)
			fprintf(stderr, "shfs filesystem not supported by the kernel\n");
		else
			perror(argv[0]);
		res = 1;
		goto out;
	}
	res = !update_mtab(mnt, long_host, *opt == ',' ? opt+1 : opt);

out:
#ifdef USE_PIPES
	close(shout[0]);
	close(shin[1]);
#endif
	free(long_host);
	free(mnt);
	exit(res);
}

